'use strict'

class DriveModeText {
    constructor(textElement, textLightness) {
        this.textElement = textElement
        this.textLightness = parseInt(textLightness)

        this.hasInitialized = false
        this.driveModesAreAvaliable = false
    }

    initialize(driveModesObject) {
        let name = driveModesObject.currentDriveMode.name
        this.hasInitialized = true
        if (name === '') {
            this.driveModesAreAvaliable = false
            this.textElement.textContent = ''
        } else {
            this.driveModesAreAvaliable = true
            this.textElement.addEventListener('mousedown', this.switchToNextDriveMode)
        }
    }

    update(text, color) {
        color = set_hsl_vals_of_rgb_str(color, -1, -1, this.textLightness)
        if (text === 'ESC & TC Off') { text = 'Off' }

        this.textElement.textContent = text
        this.textElement.style.fill = color
    }

    switchToNextDriveMode() {
        // from BeamNG.drive\ui\modules\apps\SimplePowertrainControl\app.js
        bngApi.activeObjectLua("if controller.getController('driveModes') then controller.getController('driveModes').nextDriveMode() else controller.getControllerSafe('esc').toggleESCMode() end")
    }
}

class FuelGauge {
    constructor(gaugeElement, gaugeBackgroundElement) {
        this.gaugeElement = gaugeElement
        this.gaugeBackgroundElement = gaugeBackgroundElement
        this.maximumCurveLength = gaugeElement.getTotalLength()
    }

    update(normalizedFuelValue) {
        setDashArrayProgressOfPath(this.gaugeElement, normalizedFuelValue, this.maximumCurveLength)
        setDashArrayProgressOfPath(this.gaugeBackgroundElement, 1 - normalizedFuelValue, this.maximumCurveLength)
    }
}

class GearText {
    constructor(textElement) {
        this.textElement = textElement
    }

    update(gearText) {
        let gearDisplayString = gearText
        if (gearDisplayString === 0) { gearDisplayString = 'N' }
        else if (gearDisplayString === -1) { gearDisplayString = 'R' }
        this.textElement.textContent = gearDisplayString
    }
}

class Indicators {
    constructor(leftIndicatorElement, rightIndicatorElement) {
        this.leftIndicatorElement = leftIndicatorElement
        this.rightIndicatorElement = rightIndicatorElement
    }

    update(leftIndicatorIsActive, rightIndicatorIsActive) {
        this.leftIndicatorElement.style.visibility = leftIndicatorIsActive ? 'visible' : 'hidden'
        this.rightIndicatorElement.style.visibility = rightIndicatorIsActive ? 'visible' : 'hidden'
    }

}

class LightsIcons {
    constructor(lowBeamIcon, highBeamIcon) {
        this.lowBeamIcon = lowBeamIcon
        this.highBeamIcon = highBeamIcon
    }

    update(highBeam, lowBeam) {
        let highBeamIsActive = Boolean(highBeam)
        let lowBeamIsActive = Boolean(lowBeam)
        if (highBeamIsActive) {
            this.highBeamIcon.style.visibility = 'visible'  
            this.lowBeamIcon.style.visibility = 'hidden'
        } else if (lowBeamIsActive) {
            this.lowBeamIcon.style.visibility = 'visible'
            this.highBeamIcon.style.visibility = 'hidden'
        } else {
            this.highBeamIcon.style.visibility = 'hidden'
            this.lowBeamIcon.style.visibility = 'hidden'
        }
    }

}

class ShowHideIcon {
    constructor(iconElement) {
        this.iconElement = iconElement
    }
    update(inputValue) {
        this.iconElement.style.visibility = inputValue ? 'visible' : 'hidden'
    }
}

class SpeedTexts {
    constructor(speedTextElement, unitTextElement) {
        this.speedTextElement = speedTextElement
        this.unitTextElement = unitTextElement
    }

    /**
     * @param {Number} speed - In m/s (meters per second).
     */
    update(wheelspeed, airspeed) {
        let speed = wheelspeed
        if (!wheelspeed && !isNaN(airspeed)) { speed = airspeed }
        let UiUnitsResult = UiUnits.speed(speed)
        this.speedTextElement.textContent = Math.round(UiUnitsResult.val)
        this.unitTextElement.textContent = UiUnitsResult.unit
    }
}

class TachometerArc {
    constructor(svgCurve) {
        this.svgCurve = svgCurve
        this.maximumCurveLength = svgCurve.getTotalLength()
        this.hasInitialized = false
    }

    initialize(minRpm, maxRpm) {
        this.minRpm = minRpm
        this.maxRpm = maxRpm
        this.rpmRange = maxRpm - minRpm
        this.hasInitialized = true
    }

    update(rpm) {
        rpm = Math.min(Math.max(this.minRpm, rpm), this.maxRpm)
        let normalizedRpm = (rpm - this.minRpm) / this.maxRpm
        let desiredCurveLength = this.maximumCurveLength * normalizedRpm
        let dashArray = desiredCurveLength + ',' + this.maximumCurveLength
        this.svgCurve.setAttribute('stroke-dasharray', dashArray)
    }
}

class TachometerArcMarkers {
    constructor(markersArcElement, markersWidth) {
        this.curveLength = markersArcElement.getTotalLength()
        this.markersArcElement = markersArcElement
        this.markersWidth = markersWidth
        this.hasInitialized = false
    }

    initialize(minRpm, maxRpm) {
        let rpmRange = maxRpm - minRpm
        let sectionLength = this.curveLength / (rpmRange / 1000)
        let dashLength = this.markersWidth
        let noDashLength = sectionLength - dashLength
        let startOffset = dashLength / 2

        this.markersArcElement.setAttribute('stroke-dasharray', dashLength + ' ' + noDashLength)
        this.markersArcElement.setAttribute('stroke-dashoffset', startOffset)

        if (rpmRange == 0) {
            this.markersArcElement.setAttribute('stroke-dasharray', '0 9999')
        }

        this.hasInitialized = true
    }
}

function numberToVisibility(number) {
    return number > 0 ? 'visible' : 'hidden'
}

function reInitializeUiElements(classInstanceArray) {
    classInstanceArray.forEach(classInstance => {
        if (typeof classInstance.hasInitialized !== undefined) {
            classInstance.hasInitialized = false
        } else {
            throw 'Class instance has no hasInitialized property.'
        }
    })
}

function setDashArrayProgressOfPath(path, normalizedValue, maximumCurveLength) {
    let desiredCurveLength = maximumCurveLength * normalizedValue
    let dashArray = desiredCurveLength + ',' + maximumCurveLength
    path.setAttribute('stroke-dasharray', dashArray)
}

/* ------- color functions ------- */
function rgb_str_to_rgb_values(str) {
    str = str.replace('#', '')
    let r = parseInt(str.substring(0, 2), 16)
    let g = parseInt(str.substring(2, 4), 16)
    let b = parseInt(str.substring(4, 6), 16)
    return [r, g, b]
}

function hsl_values_to_hsl_str(h, s, l) {
    return `hsl(${h}, ${s}%, ${l}%)`
}

function rgb_to_hsl(r, g, b) {
    // formulas from https://en.wikipedia.org/wiki/HSL_and_HSV
    r = r / 255
    g = g / 255
    b = b / 255

    let min = Math.min(r, g, b)
    let max = Math.max(r, g, b)
    let range = max - min

    // lightness and saturation
    let l = max * 0.5 + min * 0.5
    let s = range / (1 - Math.abs(2 * l - 1))
    if (l == 0 || l == 1) {s = 0}

    // hue
    let h6 = 0
    if (range == 0) {
        h6 = s = 0
    } else {
        if (max == r) {
            h6 = ((g - b) / range) % 6
        } else if (max == g) {
            h6 = (b - r) / range + 2
        } else if (max == b) {
            h6 = (r - g) / range + 4

        }
    }

    return [h6 * 60 , s * 100, l * 100]

}

/**
 * Use -1 on any of hue, saturation and lighness to keep the value from the color string. 
 * @param {*} str 
 * @param {*} hue 0-360 or -1
 * @param {*} saturation 0-100 or -1
 * @param {*} lightness 0-100 or -1
 * @returns {String} A string with the format `hsl(h, s% l%)`
 */
function set_hsl_vals_of_rgb_str(str, hue, saturation, lightness) {
    let colorValues = rgb_str_to_rgb_values(str)
    let newColorValues = rgb_to_hsl(...colorValues)

    if (hue >= 0) { newColorValues[0] = hue }
    if (saturation >= 0) { newColorValues[1] = saturation }
    if (lightness >= 0) { newColorValues[2] = lightness }

    return hsl_values_to_hsl_str(...newColorValues)
}


/* ------- beamng linking ------- */
angular.module('beamng.apps')
.directive('tabbyzTacho', [function () {
  return {
    template:
    '<object style="width:100%; height:100%; box-sizing:border-box;" type="image/svg+xml" data="/ui/modules/apps/tabbyzTacho/tacho.svg"></object>',
    replace: true,
    restrict: 'EA',
    link: function (scope, element, attrs) {
        let streams = ['driveModesInfo', 'electrics', 'engineInfo', 'escInfo']
        StreamsManager.add(streams)

        scope.$on('$destroy', function () {
            StreamsManager.remove(streams)
        })

        element.on('load', function () {
            let svgDoc = element[0].contentDocument
            let svgElement = element[0]
            svgElement.style.opacity = 0

            let driveModeText = new DriveModeText(svgDoc.getElementById('driveModeText'), 70)
            let fuelGauge = new FuelGauge(svgDoc.getElementById('fuelGauge'), svgDoc.getElementById('fuelGaugeBackground'))
            let gearText = new GearText(svgDoc.getElementById('gearText'))
            let pBrake = new ShowHideIcon(svgDoc.getElementById('pBrakeIcon'))
            let tachometerArc = new TachometerArc(svgDoc.getElementById('rpmCurve'))
            let tachometerArcMarkers = new TachometerArcMarkers(svgDoc.getElementById('rpmCurveMarkers'), 0.5)

            let absIcon = new ShowHideIcon(svgDoc.getElementById('absIcon'))
            let escIcon = new ShowHideIcon(svgDoc.getElementById('escIcon'))
            let tcIcon = new ShowHideIcon(svgDoc.getElementById('tcIcon'))

            let leftSignalIcon = svgDoc.getElementById('leftSignalIcon')
            let rightSignalIcon = svgDoc.getElementById('rightSignalIcon')
            let indicators = new Indicators(leftSignalIcon, rightSignalIcon)

            let lowBeamIcon = svgDoc.getElementById('lowBeamIcon')
            let highBeamIcon = svgDoc.getElementById('highBeamIcon')
            let lightIcons = new LightsIcons(lowBeamIcon, highBeamIcon)

            let speedText = svgDoc.getElementById('speedText')
            let speedUnitText = svgDoc.getElementById('speedUnitText')
            let speedTexts = new SpeedTexts(speedText, speedUnitText)

            scope.$on('streamsUpdate', function (event, streams) {
                let streamIsValid = false
                if (streams.electrics && streams.engineInfo && streams.engineInfo[1]) {
                    streamIsValid = true
                }

                if (!streamIsValid) {
                    svgElement.style.opacity = 0
                } else {
                    svgElement.style.opacity = 1
                    
                    // Logs all values that are sent to the app.
                    // console.log(streams)

                    let maxRpmDisplayed = streams.engineInfo[1] * (300 / 240)
                    if (!tachometerArc.hasInitialized) { tachometerArc.initialize(0, maxRpmDisplayed) }
                    if (!tachometerArcMarkers.hasInitialized) { tachometerArcMarkers.initialize(0, maxRpmDisplayed) }
                    if (!driveModeText.hasInitialized) { driveModeText.initialize(streams.driveModesInfo) }

                    indicators.update(streams.electrics.signal_L, streams.electrics.signal_R)
                    fuelGauge.update(streams.electrics.fuel)
                    gearText.update(streams.engineInfo[5])
                    lightIcons.update(streams.electrics.highbeam, streams.electrics.lowbeam)
                    pBrake.update(streams.electrics.parkingbrake)
                    speedTexts.update(streams.electrics.wheelspeed, streams.electrics.airspeed)
                    tachometerArc.update(streams.electrics.rpmTacho)

                    absIcon.update(streams.electrics.absActive)
                    escIcon.update(streams.electrics.escActive)
                    tcIcon.update(streams.electrics.tcsActive)

                    if (driveModeText.hasInitialized && driveModeText.driveModesAreAvaliable && streams.escInfo) {
                        let name = streams.driveModesInfo.currentDriveMode.name
                        let color = streams.escInfo.activeColor
                        driveModeText.update(name, color)
                    }
                }
            })
    
            scope.$on('VehicleFocusChanged', function (event, data) {
                reInitializeUiElements([driveModeText, tachometerArc, tachometerArcMarkers])
            })
        })
    }
  }
}])